#!/bin/sh
#
# When overlaying the sdcard over the internal /var/log
# folder, file name collision are resolved that only the file
# in the sdcard is shown, the file in the internal flash is hidden
# This script will check for such duplicates and move the file from
# the internal flash to the backup folder on the sdcard
#

BACKUP_DIR=/media/card/backup

# these variables can be overridden from the unit test
ROOTFS_MNTPOINT=${ROOTFS_MNTPOINT:-/mnt/rootfs}
CMDLINE=${CMDLINE:-/proc/cmdline}
MTAB=${MTAB:-/etc/mtab}
FSTAB=${FSTAB:-/etc/fstab}


usage() {
    echo "Usage: sdcard_trim_overlays [-x]
    -x execute, only dry-run without this option
    Will parse the fstab for union mount entries and delete those file from the
    read-only branch that also exist in the writable branch"
    exit
}

# type -p does not work in dash
if type nice > /dev/null 2>&1; then
    DO=nice
fi

collisions() {
    RO=$1
    RW=$2
    for f in `find $RO -type f`; do
        # strip $RO directory prefix
        RELPATH=${f##$RO/}
        if [ -f $RW/$RELPATH ]; then
            echo $RELPATH
        fi
    done
}

backup_file() {
    SRC=$1
    BAK_DIR=$2/$(date "+%Y-%m-%d")
    if [ ! -d $(dirname $BAK_DIR/$SRC) ]; then
        $DO mkdir -p $(dirname $BAK_DIR/$SRC)
    fi
    $DO mv $SRC $BAK_DIR/$SRC
}

rootfs_dev()
{
    sed 's/.*root=\([^ ]*\).*/\1/' $CMDLINE
}

rootfs_mounted()
{
    #
    # mountpoint(1) does not work if root device is mounted
    # twice. it stat(2) '/' and '/mnt/rootfs' and compares
    # st_dev/st_ino, if they differs it's a mointpoint
    # in our case this will not work, since both times
    # it is the root device
    # http://code.metager.de/source/xref/gnu/sysvinit/src/mountpoint.c#119
    #
    grep -q "$(rootfs_dev)\s*${ROOTFS_MNTPOINT}\s" $MTAB
}

mount_rootfs()
{
    ROOTDEV="$(rootfs_dev)"
    if ! $(expr ${ROOTDEV} : '.*\(mtdblock\).*' > /dev/null); then
        echo "ERROR: invalid root device $ROOTDEV" >&2
        return 1
    fi

    ROOTFS_TYPE=$(sed 's/.*rootfstype=\([^ ]*\).*/\1/' $CMDLINE)
    if [ -n $ROOTFS_TYPE -a $ROOTFS_TYPE != 'jffs2' ]; then
        echo "ERROR: unsupported filesystem $ROOTFS_TYPE" >&2
        return 2
    fi

    if rootfs_mounted; then
        echo "already mounted"
        return 0
    fi

    if [ ! -d ${ROOTFS_MNTPOINT} ]; then
        $DO mkdir -p ${ROOTFS_MNTPOINT}
    fi

    $DO mount -t $ROOTFS_TYPE $ROOTDEV $ROOTFS_MNTPOINT
}

umount_rootfs()
{
    if rootfs_mounted; then
        umount $ROOTFS_MNTPOINT
    fi
}

parse_overlays()
{
    #
    # technically unions consists of branches(br) of
    # directories, where the first directory is implicitly
    # writable and those that follow are read-only.
    # permission can be overridden, as is in our case
    # br=writable[=rw]:readable-dir[=ro]
    # NOTE: make sure to emit RO dir before RW directory
    #
    awk '/^aufs/ {print $4}' $FSTAB | \
        sed -n -e 's/.*br:\([[:graph:]]*\)=rw:\([[:graph:]]*\)=ro/\2 \1/p' \
        -e 's/.*br:\([[:graph:]]*\)=ro:\([[:graph:]]*\)=rw/\1 \2/p' \
        -e 's/.*br:\([[:graph:]]*\):\([[:graph:]]*\)/\2 \1/p'
}

# main
if [ -n "$UNITTEST" ]; then
    # exit would abort unittest, but quit here
    return
fi

if ! mountpoint /media/card; then
    echo "sdcard not mounted"
    exit
fi

trap umount_rootfs EXIT SIGHUP SIGINT SIGQUIT SIGABRT SIGKILL SIGTERM

if [ $# -ne 0 -a "x$1" != "x-x" ]; then
    usage
fi

if ! mount_rootfs; then
    return 255
fi

if [ "x$1" != "x-x" ]; then
    # must mount before
    echo "-- dry-run"
    DO="echo --"
fi

parse_overlays | while read RO RW; do
# access it via the double mounted rootfs
RO=$ROOTFS_MNTPOINT$RO
for REL_PATH in `collisions $RO $RW`; do
    echo "$RO> backing up $REL_PATH"
    backup_file $RO/$REL_PATH $BACKUP_DIR
done
done
